"""
This file contains the color scheme for some widgets and backgrounds of
TriFusion
"""

# Main colors for general usage in the application. They redirect to _blue and
# _red in app.py
c_app_blue = (0.216, 0.67, 0.784, 1)
c_app_red = (1., 0.39, 0.39, 1.)

# Main window background
c_main_background = (.08, .08, .08, 1)

# Sidepanel background
c_iner_side_panel_background = (0.1, 0.1, 0.1, .9)
c_outer_side_panel_background = c_iner_side_panel_background

# Sidepanel Caroussel headers
c_side_panel_headers = (.3, .3, .3, 1)

# Light boxes
c_light_box = (.2, .2, .2, 1)

# Dark boxes
c_dark_box = (0, 0, 0, 0)

# Orthology sidepanel
c_orto_sidepanel = (.2, .2, .2, .5)

# Statistics side panel
c_stats_sidepanel = (0.1, 0.1, 0.1, .7)

# Popup background color
c_popup_background = c_main_background
c_popup_border = (0.216, 0.67, 0.784, 1)

# Filechoosers headers
c_filechooser_headers = (.2, .2, .2, .9)

# Dropdown backgrounds
c_dropdown_background = (.15, .15, .15, 1)
